#if !defined(_BWT_)
#define _BWT_



#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>

#if ( INT_MAX == 32767 )
#define BLOCK_SIZE 20000
#else
#define BLOCK_SIZE 40000
#endif


//um qsort von innerhalb der Klasse ansprechen zu koennen :
typedef int (__cdecl *GENERICCOMPAREFN)(const void *elem1 , const void *elem2);

//length = anzahl der momentan im buffer eingelesenen Bytes
//buffer = zu transformierende Daten
//indices = Indizes der verschobenen Zeilen (fuer Sortieren benoetigt)
long			length;
unsigned char	buffer[BLOCK_SIZE+1];
int				indices[BLOCK_SIZE+1];


class BWT
{
	public :

		BWT(void){};
//		int	BWT_Encode(char *f1 , char *f2);
//		int	BWT_Decode(char *f1 , char *f2);
		int	BWT_Encode(FILE *fin , FILE *fout);
		int	BWT_Decode(FILE *fin , FILE *fout);


	protected :

		static int __cdecl row_compare( const unsigned int *i1, const unsigned int *i2 );
};





//#################################################################
int __cdecl BWT::row_compare( const unsigned int *i1, const unsigned int *i2 )
{
	unsigned int l1 = (unsigned int) ( length - *i1 );
	unsigned int l2 = (unsigned int) ( length - *i2 );
	int result;
	result = memcmp( buffer + *i1 , buffer + *i2 , l1 < l2 ? l1 : l2 );

	if (result == 0)
		return l2 - l1;
	else
	return result;
}



//####################################################################
//int BWT::BWT_Encode(char *f1 , char *f2)
int	BWT::BWT_Encode(FILE *fin , FILE *fout)
{
//    freopen( f1, "rb", stdin );
//    freopen( f2, "wb", stdout );

	//Kompresion von maximal BLOCK_SIZE Bytes
	//und zwar bis fread 0 liest
	for ( ; ; )
	{
		length = fread((char *)buffer , 1 , BLOCK_SIZE , fin);
		if (length == 0)
			break;
//		fprintf( stderr, "Performing BWT on %ld bytes\n", length );
		long l = length + 1;
		fwrite( (char *) &l, 1, sizeof( long ), fout );

		//Lexikographisches Sortieren der Zeilen (bzw. deren Indizes)
		int i;
		for (i=0; i<=length; i++)
			indices[i] = i;

		qsort( indices , (int)(length+1) , sizeof(int) ,
				(GENERICCOMPAREFN) row_compare );


		//Letzte Spalte L in den OutputStream schreiben
		//dazu noch die Position des end-of-buffer Zeichens
		//in L finden sowie das erste Zeichen
	
		long first;
		long last;
		for (i=0; i<=length; i++)
		{
			if (indices[i] == 1)
				first = i;
			if (indices[i] == 0)
			{
				last = i;
				fputc( '?', fout );
			}
			else
				fputc(buffer[indices[i]-1], fout);
		}

		fwrite( (char *) &first , 1 , sizeof(long) , fout );
		fwrite( (char *) &last , 1 , sizeof(long) , fout );
	}

	return 0;
}








//##################################################################
//int BWT::BWT_Decode(char *f1 , char *f2)
int	BWT::BWT_Decode(FILE *fin , FILE *fout)
{
	unsigned int T[BLOCK_SIZE+1];
	long buflen;
	unsigned int Count[257];
	unsigned int RunningTotal[257];
	unsigned long first , last;

//	freopen(f1 , "rb" , stdin);
//	freopen(f2 , "wb" , stdout);

	for ( ; ; )
	{
		if ( fread((char *) &buflen , sizeof(long) , 1 , fin) == 0 )
			break;
		if (buflen > (BLOCK_SIZE+1) )
		{
			fprintf( stderr, "Buffer overflow!\n" );
			abort();
		}
		if ( fread((char *)buffer , 1 , (size_t) buflen , fin ) != (size_t) buflen)
		{
			fprintf( stderr, "Error reading data\n" );
			abort();
		}

		fread((char *)&first , sizeof(long) , 1 , fin);
		fread((char *)&last , sizeof(long) , 1 , fin);

		//Die Position der Zeichen im Output-buffer bestimmen
		unsigned int i;
		for (i=0; i<257; i++)
			Count[i] = 0;
		for (i=0; i<(unsigned int)buflen; i++)
			if ( i == last )
				Count[256]++;
			else
				Count[buffer[i]]++;
		int sum = 0;
		for (i=0; i<257; i++)
		{
			RunningTotal[i] = sum;
			sum += Count[ i ];
			Count[ i ] = 0;
		}

		//schliesslich den Transformationsvektor T[] bauen
		for (i=0; i<(unsigned int)buflen; i++)
		{
			int index;
			if (i == last)
				index = 256;
			else
				index = buffer[i];
			T[ Count[index] + RunningTotal[index] ] = i;
			Count[index]++;
		}

		//und schreiben
		unsigned int j;
		i = (unsigned int) first;
		for (j=0; j<(unsigned int)(buflen-1); j++)
		{
			putc(buffer[i] , fout);
			i = T[i];
		}
	}

	return 1;
}




#endif